@@ -6,7 +6,7 @@
     | modification, are permitted provided that the conditions mentioned |
     | in the accompanying LICENSE file are met.                          |
     +--------------------------------------------------------------------+
-    | Copyright (c) 2004-2013, Michael Wallner <mike@php.net>            |
+    | Copyright (c) 2004-2014, Michael Wallner <mike@php.net>            |
     +--------------------------------------------------------------------+
 */
 
@@ -60,9 +60,6 @@ static zend_module_dep http_module_deps[] = {
 #endif
 #ifdef PHP_HTTP_HAVE_JSON
 	ZEND_MOD_REQUIRED("json")
-#endif
-#ifdef PHP_HTTP_HAVE_EVENT
-	ZEND_MOD_CONFLICTS("event")
 #endif
 	{NULL, NULL, NULL, 0}
 };
@@ -190,9 +187,6 @@ PHP_RINIT_FUNCTION(http)
 {
 	if (0
 	|| SUCCESS != PHP_RINIT_CALL(http_env)
-#if PHP_HTTP_HAVE_CURL && PHP_HTTP_HAVE_EVENT
-	|| SUCCESS != PHP_RINIT_CALL(http_client_curl)
-#endif
 	) {
 		return FAILURE;
 	}
@@ -203,9 +197,6 @@ PHP_RINIT_FUNCTION(http)
 PHP_RSHUTDOWN_FUNCTION(http)
 {
 	if (0
-#if PHP_HTTP_HAVE_CURL && PHP_HTTP_HAVE_EVENT
-	|| SUCCESS != PHP_RSHUTDOWN_CALL(http_client_curl)
-#endif
 	|| SUCCESS != PHP_RSHUTDOWN_CALL(http_env)
 	) {
 		return FAILURE;
